(*
author:	Alexander Jahraus
date:	July 1, 2007
version: 1.0

Dieses Script ordnet bis zu 16 Finder-Fenster auf dem
Desktop so an, dass sie nicht berlappen und in einem
Gitter tabellenartig angeordnet werden.

Dabei werden die Fenster bewegt und ihre Gre wird
verndert. Die neue Position fr jedes Fenster wird
so gewhlt, dass das Fenster einen minimalen Weg zurck-
legen muss.

Die Bewegung und nderung der Fenstergre ist animiert;
die Anzahl der Animationsschritte pro Bewegung kann 
unten in animationSteps angepasst werden.
*)

set displayHighWidth to my displays_info()
set screenHeight to (item 2 of displayHighWidth) --HIER HHE DES MONITORS IN PIXEL EINTRAGEN
set DockHeight to (do shell script ("defaults read com.apple.dock tilesize")) as integer --HIER HHE DES DOCK IN PIXEL
global screenWidth
set screenWidth to (item 1 of displayHighWidth) --HIER BREITE DES BILDSCHIRMS IN PIXEL


global heightWithDock --Hhe der Arbeitsflche mit Dock an
global heightWithoutDock --und Dock aus
global menuBarHeight --Hhe der Menleiste (=22 plus Finder-Fenstertitel)
global ursprung --oberer linker Punkt der Arbeitsflche
global animationSteps --Anzahl der Animationsschritte
global unitX --Breite jedes Slots
global unitY --Hhe jedes Slots


set menuBarHeight to 44
set ursprung to {0, menuBarHeight} -- Punkt links oben auf dem Desktop unter Menleiste
set heightWithDock to screenHeight - menuBarHeight - DockHeight
set heightWithoutDock to screenHeight - menuBarHeight
set animationSteps to 9 --larger number --> smoother animation but longer calculation;
--						numbers over 17 just slow down calculation and don't have any effect on the animation


tell application "Finder"
	set windowCount to count every window --make windowList-------------------------------------
	if not windowCount is 0 then
		set windowList to {1} -- Liste mit einem Dummy-Element
		repeat with i from 1 to windowCount
			set item i of windowList to (a reference to window i)
			if i < windowCount then
				set windowList to windowList & 1 --Liste um eins verlngern, auer am Schluss
			end if
		end repeat
		set newBoundsList to my makeNewBounds(windowList, my calculateGrid(windowCount))
		my itsMagic(newBoundsList)
	end if
end tell


on getExpFragmentation(distance, parts) -------------------------------------------------------
	(*	this function takes an two integers and returns an exponential fragmentation
	of [distance] in [parts] parts as a list so that the fragments are like
	{x, 2x, ... 2^(parts div 2)x + z, ... 2x, x} and the sum of the list is [distance]
	*)
	if distance < 10 then
		return {distance} -- unter 10px keine Segmentierung
	end if
	
	if parts < 3 then
		set parts to 3 -- mind. 3 Segmente
	end if
	
	if parts > 19 then
		set parts to 19 -- 19 ist die maximale Anzahl der Teile fr 2264px
	end if
	
	if parts mod 2 = 0 then
		set parts to parts + 1 -- make number of parts uneven
	end if
	
	set unit to 0
	repeat until unit > 0 --unit darf nicht null sein
		set X to 1
		set mitte to parts div 2 - 1
		repeat mitte times
			set X to 2 * X + 1
		end repeat
		set Y to 3 * X + 1 -- unit is Y times in distance
		set unit to distance div Y
		if unit = 0 then
			set parts to parts - 2 --notfalls Segmentierung verringern
		end if
	end repeat
	
	set modulo to distance mod Y
	set leftPart to {}
	
	repeat with i from 0 to mitte
		set currentValue to unit * (2 ^ i) as integer
		set leftPart to leftPart & currentValue
	end repeat
	return leftPart & unit * (2 ^ (mitte + 1)) + modulo & reverse of leftPart --rightPart
end getExpFragmentation

--test: log getExpFragmentation(1532, 17)

on maximum(listOfNumbers) --berechnet das Maximum der Zahlen einer Liste------------------------------
	set max to 0
	repeat with i in listOfNumbers
		if i > max then
			set max to i
		end if
	end repeat
	return max
end maximum

on minimum(listOfNumbers) --berechnet das Minimum der Zahlen einer Liste------------------------------
	set min to first item in listOfNumbers
	repeat with i in rest of listOfNumbers
		if i < min then
			set min to i
		end if
	end repeat
	return min
end minimum

on qabstand(p1, p2) -- berechnet den quadratischen Abstand zwischen zwei Punkten in pxl^2------------
	set xdiff to ((first item of p1) - (first item of p2))
	set ydiff to ((second item of p1) - (second item of p2))
	return xdiff * xdiff + ydiff * ydiff
end qabstand

on nummer(element, liste) ------------------------------------------------------------------------------------
	(* returns index of first appearance of an integer in a list
		or zero if element doesn't occur in list *)
	repeat with i from 1 to length of liste
		if (item i of liste) as integer = element as integer then
			return i
		end if
	end repeat
	return 0
end nummer

on calculateGrid(n) -- berechnet das Ausrichtungsgitter fr n Fenster----------------------------------
	set width to screenWidth
	set height to heightWithDock
	
	set cols to n div 5 + 2 --Anzahl Spalten
	if n is in {13, 14} then
		set cols to 5
	end if
	
	if n < 3 then --Anzahl Zeilen
		set rows to 1
	else if n < 7 then
		set rows to 2
	else
		set rows to 3
	end if
	
	if n  16 then -- bei 16 Fenstern: 4x4
		set n to 16 --maximal 16 Fenster ausrichten: dann Dock ausblenden
		tell application "System Events"
			keystroke "d" using {command down, option down} -- hide the Dock
		end tell
		set height to heightWithoutDock
		set cols to 4
		set rows to 4
	end if
	
	set unitX to width div cols --breite eines Fensters
	set unitY to height div rows --Hhe eines Fensters
	
	set grid to {}
	repeat with i from 0 to rows - 1
		repeat with j from 0 to cols - 1
			set grid to grid & {{coords:{j * (unitX), i * (unitY) + menuBarHeight}, occupied:false}}
		end repeat
	end repeat
	return grid
end calculateGrid -------------------------------------------------------------------------------

on makeNewBounds(windowList, grid) ----------------------------------------------------------
	tell application "Finder"
		(* hier ist der Dummy fr den Record fr die Fenster-Resize-Operation.
	win ist das Fenster,
	oldPos / newPos - Position der linken oberen Ecke,
	oldLRC / newLRC - Position der rechten unteren Ecke *)
		set newBounds to {}
		--um bei mehr als 16 Fenstern keine Endlosschleife zu erzeugen:
		set maxWindows to my minimum({16, length of windowList})
		log "maxWindows: " & maxWindows
		repeat with w in (items 1 thru maxWindows of windowList)
			--ermitteln, welchem Gitterpunkt dieses Fenster am nchsten ist
			--dazu erst Liste der Abstnde zu Gitterpunkten aufbauen
			set oldBounds to ((bounds of w) as list)
			set oldPosition to (items 1 thru 2 of oldBounds)
			set distances to {}
			repeat with j from 1 to length of grid
				set distances to distances & 
					my qabstand(oldPosition, 
						coords of item j of grid)
			end repeat
			--			log distances
			--jetzt den minimalen Abstand berechnen,
			set minDistance to my minimum(distances)
			--			log minDistance
			--seine Nummer in der Liste der Abstnde ermitteln
			--und den entsprechenden Gitterpunkt finden
			set minDisNr to my nummer(minDistance, distances)
			--			log minDisNr
			repeat while (occupied of item minDisNr of grid) --falls Gitter-Slot schon belegt, whle den nchsten
				--				log (occupied of item minDisNr of grid)
				set minDisNr to minDisNr + 1
				if minDisNr > length of grid then
					set minDisNr to 1
				end if
			end repeat
			-- alles in den Record schreiben, diesen an die Liste newBounds anhngen
			set newBounds to newBounds & 
				{{win:w, oldPos:oldPosition, newPos:coords of item minDisNr of grid, oldLRC:items 3 thru 4 of oldBounds, newLRC:{(first item of coords of item minDisNr of grid) + unitX, (second item of coords of item minDisNr of grid) + unitY}}}
			
			set occupied of item minDisNr of grid to true --Slot als belegt markieren
		end repeat
	end tell
	return newBounds
end makeNewBounds

on abs(X) --Absolutbetrag von x (x integer oder x real)
	if X < 0 then
		return 0 - X
	else
		return X
	end if
end abs

on itsMagic(map) --verschiebt die Fenster und ndert deren Gre (animiert)---------------------------
	tell application "Finder"
		repeat with i from 1 to length of map -- map ist nur eine kurze Schreibweise fr die Liste newBounds
			set eintrag to item i of map
			set fenster to win of item i of map
			set theBounds to (bounds of fenster as list)
			
			set x1 to first item of theBounds --(x1,y1) ist der obere linke Punkt
			set x2 to item 3 of theBounds --(x2,y2) ist der untere rechte Punkt
			set Y1 to second item of theBounds
			set Y2 to item 4 of theBounds
			
			set dx1 to ((first item of newPos of eintrag) - (x1)) --(dx1,dy1) ist der Verschiebungsvektor fr oben links
			set dx2 to ((first item of newLRC of eintrag) - (x2)) --(dx2,dy2) ist der Verschiebungsvektor fr unten rechts
			set dy1 to ((second item of newPos of eintrag) - (Y1))
			set dy2 to ((second item of newLRC of eintrag) - (Y2))
			
			-- Maximum der Absolutbetrge der Vektorkomponenten finden
			set maxAbs to my maximum({my abs(dx1), my abs(dx2), my abs(dy1), my abs(dy2)})
			set steps to my getExpFragmentation(maxAbs, animationSteps) --diesen maximalen Abstand in [animationSteps] Schritte teilen
			
			set sums to {}
			set currentSum to 0 --Teilsummen der Schritte berechnen
			repeat with s in steps
				set currentSum to currentSum + s
				set sums to sums & currentSum
			end repeat
			
			set animation to {} -- animation ist Liste von bounds-Quadrupel zum Durchlaufen der Animation
			repeat with s in sums
				if maxAbs = 0 then
					set p to 0
				else
					set p to s / maxAbs --Prozent der schon durchlaufenen Strecke
				end if
				set animation to animation & {{x1 + (p * dx1) as integer, Y1 + (p * dy1) as integer, x2 + (p * dx2) as integer, Y2 + (p * dy2) as integer}}
			end repeat
			
			repeat with viereck in animation
				set bounds of fenster to viereck
			end repeat
		end repeat
	end tell
end itsMagic

on displays_info()
	set plistFolderPath to (path to preferences folder as string) & "ByHost:"
	tell application "Finder" to set plistName to (name of files of folder plistFolderPath whose name contains "com.apple.windowserver") as string
	set plistPath to plistFolderPath & plistName
	set j to 1
	set theColors to {"Greyscale", "256", "Thousands", "Millions"}
	tell application "Image Events" to set count_of_displays to count of displays
	repeat
		tell application "System Events" to set displaySet to value of property list item j of property list item "DisplaySets" of property list file plistPath
		try
			repeat with i from 1 to count_of_displays
				set {expID, theWidth, theHeight, theDepth, theRate, theOriginx, theOriginy} to {|DisplayID| of (item i of displaySet), |Width| of (item i of displaySet), |Height| of (item i of displaySet), |Depth| of (item i of displaySet), |RefreshRate16.16| of (item i of displaySet), |OriginX| of (item i of displaySet), |OriginY| of (item i of displaySet)}
				set exp_to_string to words of (expID as string)
				set stringID to (character 1 of (item 1 of exp_to_string)) & (characters 3 thru -2 of (item 1 of exp_to_string)) as string
				if (count of stringID) < ((item 3 of exp_to_string) + 1) then
					repeat with i from (count of stringID) to (item 3 of exp_to_string)
						set stringID to stringID & "0"
					end repeat
				end if
				set thisColorDepth to item theDepth of theColors -- change depth to a color value
			end repeat
			exit repeat
		on error
			set j to j + 1
		end try
	end repeat
	return {theWidth as integer, theHeight as integer}
end displays_info